/* ------------------------------------------------------------ *\
 |                                                              |
 | Module  : m4mqr3.c                                           |
 |                                                              |
 | Description  : Sample Exit for use with the MQSeries Link    |
 |                SAP R/3. This sample is given to illustrate   |
 |                transformation of -                           |
 |                  an inbound message which is not in the      |
 |                SAP R/3 IDOC format to an IDOC form as        |
 |                required by using the Mercator Platform API   |
 |                to call a map that will take the application  |
 |                data and map it into IDoc format.             |
 |                  an outbound message in IDOC format as       |
 |                produced by some function into a user-defined |
 |                format suitable for processing by a non-SAP   |
 |                system (in fact the reverse of the inbound    |
 |                transformation). The user-defined formats are |
 |                defined in the sample header file, m4mqr3b.h. |
 |                                                              |
 |                It also illustrates the use of MQ calls       |
 |                within an exit.                               |
 |                                                              |
 |                Four entry points should be provided by an    |
 |                exit viz.                                     |
 |                 InitialiseEntry: this is called at start     |
 |                   of the server. This must be the first      |
 |                   'exported' entry point for the dll/sl.     |
 |                 TerminateEntry: called at termination of     |
 |                   the server.                                |
 |                 ExecuteEntry: called after a message is      |
 |                   retrieved from a queue (inbound server)    |
 |                   and before a message is placed on the      |
 |                   destination queue (Outbound server).       |
 |                 ReturnEntry: called after all processing     |
 |                   has been completed for the message dealt   |
 |                   with in the previous invication of the     |
 |                   Execute function.                          |
 |                                                              |
 | Notes  : When compiling this exit on Windows/NT the          |
 |          compiler variable SMQ_NT should be defined.         |
 |                                                              |
 |          This code was adapted from the sample code given    |
 |          with the MQSeries Link for R/3 (Windows NT) product.|
 |          The two functions that have undergone code changes  |
 |          are the transInbound and transOutbound functions.   |
 |          This is a development sample used to aid others in  |
 |          how the Mercator Platform API may be embeded into   |
 |          a User Exit for the Link software.                  |
 |                                                              |
 |          This is NOT production code it is for demonstration |
 |          purposes ONLY.                                      |
 |                                                              |
\* ------------------------------------------------------------ */

/* ------------------------------------------------------------ *\
 | Include some 'standard' C header files                       |
\* ------------------------------------------------------------ */
#include <windows.h>
#include <runmerc.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#if defined(SMQ_NT)
  #include <process.h>
#endif

/* ------------------------------------------------------------ *\
 | Include the header files for both MQSeries and the Link      |
\* ------------------------------------------------------------ */
#include "cmqc.h"
#include "smqc.h"

/* ------------------------------------------------------------ *\
 | Include necessary SAP header files                           |
\* ------------------------------------------------------------ */

/* ------------------------------------------------------------ *\
 | Include our own header files                                 |
\* ------------------------------------------------------------ */
#include "m4mqr3a.h"
#include "m4mqr3b.h"

/* ------------------------------------------------------------ *\
 | Include Mercator Platform API header file                    |
\* ------------------------------------------------------------ */
#include "runmerc.h"

/* ------------------------------------------------------------ *\
 | Define prototypes for the exit entry points. There are       |
 | four of these viz.                                           |
 |   Initialise                                                 |
 |   Execute                                                    |
 |   Return                                                     |
 |   Terminate                                                  |
 | The typedefs for the entry points are contained in smqc.h.   |
\* ------------------------------------------------------------ */
smqUserExitInitialise Initialise ;
smqUserExitExecute Execute ;
smqUserExitReturn Return ;
smqUserExitTerminate Terminate ;

/* ------------------------------------------------------------ *\
 | define some default structures.                              |
\* ------------------------------------------------------------ */
static MQSAPH DefHeader = {MQSAPH_DEFAULT} ;



/**************************************************************/
/*                                                            */
/* Routine  : Initialise                                      */
/*                                                            */
/* Description  : Initialisation routine for the in/outbound. */
/*                transformation sample exit. It is called by */
/*                server and is responsible for returning the */
/*                addresses of the other entry points. It may */
/*                also perform any additional processing      */
/*                needed for subsequent calls to the exit.    */
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure.                                 */
/*               pExitInitialise - pointer to an SMQEXIT_INIT */
/*                 structure.                                 */
/*                                                            */
/**************************************************************/
#define EXIT_FUNCTION "Initialise"
void MQENTRY Initialise(PSMQEXIT_SETUP pExitSetup,
                        PSMQEXIT_INIT  pExitInitialise)
{
  /* ------------------------------------------------------ *\
   | Declare our local variables                            |
  \* ------------------------------------------------------ */
  MQCHAR48  LogQName = {LOG_QUEUE_NAME} ;
                               /* the name of our log queue   */
  MQHCONN   QMgrHandle = MQHC_UNUSABLE_HCONN ;
                               /* connection handle           */
  MQHOBJ    LogQHandle = MQHO_UNUSABLE_HOBJ ;
                               /* handle for log queue        */
  MQLONG    CompCode ;         /* Completion/reason codes for */
  MQLONG    ReasonCode ;       /* MQ calls.                   */
  MQLONG        OpenOptions ;
  MQOD      ObjectDesc = {MQOD_DEFAULT} ;
  MQMD      DefMQMD = {MQMD_DEFAULT} ;
  MQPMO     DefMQPMO = {MQPMO_DEFAULT} ;
  TransLog  DefTLog = {TransLog_DEFAULT} ;
  pTransLog pLogData = NULL ;  /* pointer to our control      */
                               /* block                       */
  pLogBuffer pTBuffer ;        /* pointer to the buffer we    */
                               /* acquire to hold our log info*/
          ;
  /* ------------------------------------------------------ *\
   | Set a default good result code                         |
  \* ------------------------------------------------------ */
  pExitInitialise -> Result = SMQRC_OK ;

  /* ------------------------------------------------------ *\
   | First we check the input structures are things we can  |
   | handle i.e. the versions of the structures are those we|
   | were compiled with. If we don't recognise the versions |
   | then we return immediately with TERMINATE_EXIT. The    |
   | exit handler will not then call the Execute and Return |
   | functions (but will call the terminate function).      |
  \* ------------------------------------------------------ */
  if ((pExitSetup -> Version != SMQEXIT_SETUP_VERSION) ||
      (pExitInitialise -> Version != SMQEXIT_INIT_VERSION))
  {
    pExitInitialise -> Result = SMQRC_TERMINATE_EXIT ;
    return ;
  }
  /* ------------------------------------------------------ *\
   | At a minimum we must return the addresses of the other |
   | entry points for the exit. The init structure contains |
   | the fields for these pointers.                         |
  \* ------------------------------------------------------ */
  pExitInitialise -> ExecEntryPoint = (void *)Execute ;
  pExitInitialise -> RetnEntryPoint = (void *)Return ;
  pExitInitialise -> TermEntryPoint = (void *)Terminate ;

  /* ------------------------------------------------------ *\
   | Save the QMgr connection handle in our local storage.  |
  \* ------------------------------------------------------ */
  QMgrHandle = pExitInitialise -> MQHandle ;

  /* ------------------------------------------------------ *\
   | Allocate ourselves a control block which we use to pass|
   | info between calls to the exit. We use the ExitHandle  |
   | field of the setup structure to contain the address of |
   | block. Failure to allocate the required storage is     |
   | regarded as fatal and we terminate the server.         |
  \* ------------------------------------------------------ */
  if ((pLogData = malloc(sizeof(TransLog))) == NULL)
  {
    pExitSetup -> ExitOptions = TRANS_STORAGE_ERROR ;
        pExitInitialise -> Result = SMQRC_TERMINATE_SERVER ;
        goto INITEXIT ;
  }
  pExitSetup -> ExitHandle = (MQLONG) pLogData ;
  memcpy(pLogData,
         &DefTLog,
             sizeof(TransLog)) ;
  pLogData -> QMgrHandle = QMgrHandle ;

  /* ------------------------------------------------------ *\
   | Now attempt to open our log queue. We open it for      |
   | output and inquire. Again failure here is regarded as  |
   | fatal and we ask the server to terminate.              |
  \* ------------------------------------------------------ */
  OpenOptions = MQOO_OUTPUT +
                    MQOO_FAIL_IF_QUIESCING +
                    MQOO_INQUIRE ;
  strncpy(ObjectDesc.ObjectName,
              LogQName,
                  MQ_Q_NAME_LENGTH) ;
  MQOPEN(QMgrHandle,
             &ObjectDesc,
                 OpenOptions,
             &LogQHandle,
                 &CompCode,
                 &ReasonCode) ;
  if (CompCode != MQCC_OK)
  {
    pExitSetup -> ExitOptions = TRANS_OPEN_ERROR ;
        pExitInitialise -> Result = SMQRC_TERMINATE_SERVER ;
        goto INITEXIT ;
  }

  /************************************************************/
  /* Save the handle for the logQ in our control block. Also  */
  /* take note of the direction.                              */
  /************************************************************/
  pLogData -> LogQHandle = LogQHandle ;
  pLogData -> Direction =
      pExitInitialise -> Direction ;

  /************************************************************/
  /* Now allocate space for an MQMD, our buffer for the log   */
  /* queue and for a put options structure. We allocate the   */
  /* storage in a single block. Once more failure here is     */
  /* regarded as fatal.                                       */
  /************************************************************/
  if ((pLogData -> pLogMQMD = malloc(sizeof(MQMD) +
                                         sizeof(MQPMO) +
                                     sizeof(LogBuffer)))
                                == NULL)
  {
        pExitSetup -> ExitOptions = TRANS_STORAGE_ERROR ;
        pExitInitialise -> Result = SMQRC_TERMINATE_SERVER ;
        goto INITEXIT ;
  }

  /***********************************************************/
  /* Now initialise the MQMD with the default values.        */
  /***********************************************************/
  memcpy(pLogData -> pLogMQMD,
                 &DefMQMD,
                 sizeof(MQMD)) ;

  /***********************************************************/
  /* Step over our newly initialised MQMD and set up our put */
  /* message options structure. Note that we set the pmo to  */
  /* no syncpoint.                                           */
  /***********************************************************/
  pLogData -> pLogMQPMO =
                  (PMQPMO)((char *)(pLogData -> pLogMQMD)
                                + sizeof(MQMD)) ;
  memcpy(pLogData -> pLogMQPMO,
         &DefMQPMO,
                 sizeof(MQPMO)) ;
  pLogData -> pLogMQPMO -> Options = MQPMO_NO_SYNCPOINT +
                                         MQPMO_FAIL_IF_QUIESCING ;
  pLogData -> pBuffer =
                  (pLogBuffer)((char *)(pLogData -> pLogMQPMO)
                                                                + sizeof(MQPMO)) ;

  /***********************************************************/
  /* Now check the direction which is contained in the init  */
  /* structure. If it's invalid then we terminate the server.*/
  /* Of course this should never happen.                     */
  /***********************************************************/
  pTBuffer = pLogData -> pBuffer ;

  switch (pLogData -> Direction)
  {
    case SMQEXIT_INBOUND:
      memcpy(pTBuffer -> InOut,
                 MSG_IN,
             4) ;
      break ;
    case SMQEXIT_OUTBOUND:
      memcpy(pTBuffer -> InOut,
                 MSG_OUT,
             4) ;
      break ;
    default:
      pExitSetup -> ExitOptions = TRANS_INV_DIRECTION ;
          pExitInitialise -> Result = SMQRC_TERMINATE_SERVER ;
      goto INITEXIT ;
      break ;
  }


  /************************************************************/
  /* Now get the process id for this instance and store it in */
  /* our log buffer.                                          */
  /************************************************************/
#if defined(SMQ_NT)
  pTBuffer -> ProcessId = _getpid() ;
#else
  pTBuffer -> ProcessId = getpid() ;
#endif

  /************************************************************/
  /* All paths through this function go through here to       */
  /* us to perform tracing etc.  .                            */
  /************************************************************/
INITEXIT:
  /************************************************************/
  /* If we successfully initialised then record this in our   */
  /* 'log' queue.                                             */
  /************************************************************/
  if (pExitInitialise -> Result == SMQRC_OK)
  {
    pTBuffer -> Result = SMQRC_OK ;
    memcpy(pTBuffer -> Phase,
               EXIT_FUNCTION,
               12) ;
        MQPUT(pLogData -> QMgrHandle,
              pLogData -> LogQHandle,
              pLogData -> pLogMQMD,
                  pLogData -> pLogMQPMO,
                  sizeof(LogBuffer),
                  pLogData -> pBuffer,
                  &CompCode,
                  &ReasonCode) ;
    if (CompCode != MQCC_OK)
    {
      pExitSetup -> ExitOptions = ReasonCode ;
      pTBuffer -> Result = SMQRC_TERMINATE_SERVER ;
      pExitInitialise -> Result = SMQRC_TERMINATE_SERVER ;
    }
  }

  /************************************************************/
  /* Now return to the exit handler.                          */
  /************************************************************/
  return ;
}
#undef EXIT_FUNCTION

/**************************************************************/
/*                                                            */
/* Routine  : Terminate                                       */
/*                                                            */
/* Description  : Termination routine for sample exit. This   */
/*                is called at termination of the server. It  */
/*                is responsible for performing any cleanup   */
/*                of resources acquired during previous exit  */
/*                processing.                                 */
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure.                                 */
/*               ExitTerminate - The reason for termination.  */
/*                                                            */
/**************************************************************/
#define EXIT_FUNCTION "Terminate   "
void MQENTRY Terminate(PSMQEXIT_SETUP pExitSetup,
                       SMQEXIT_REQUEST ExitTerminate)
{

  pTransLog pLogData ;
  MQLONG    CompCode ;
  MQLONG    ReasonCode ;
  pLogBuffer pTBuffer ;

  /************************************************************/
  /* We attempt to clean up any resources we have acquired    */
  /* in this incarnation. If the pointer to our control block */
  /* is null then we have nothing to clean up.                */
  /************************************************************/
  pLogData = (pTransLog)(pExitSetup -> ExitHandle) ;
  if (pLogData != NULL)
  {
        /**********************************************************/
        /* Check whether or not we have storage to free. If we    */
        /* have then this means we also have access to the 'log'  */
        /* queue and we can record our demise and then close the  */
    /* log queue.                                             */
        /**********************************************************/
        if (pLogData -> pLogMQMD != NULL)
        {
      pTBuffer = pLogData -> pBuffer ;
          memcpy(pTBuffer -> Phase,
                 EXIT_FUNCTION,
                         12) ;
          pTBuffer -> Result = ExitTerminate ;
          MQPUT(pLogData -> QMgrHandle,
                    pLogData -> LogQHandle,
                        pLogData -> pLogMQMD,
                        pLogData -> pLogMQPMO,
                        sizeof(LogBuffer),
                        pTBuffer,
                        &CompCode,
                        &ReasonCode) ;
      free(pLogData -> pLogMQMD) ;
          MQCLOSE(pLogData -> QMgrHandle,
                      &(pLogData -> LogQHandle),
                  MQCO_NONE,
                          &CompCode,
              &ReasonCode) ;
        }

        /**********************************************************/
        /* Finally free our control block structure.              */
        /**********************************************************/
        free(pLogData) ;
  }

  /************************************************************/
  /* Return to the exit handler.                              */
  /************************************************************/
  return ;
}
#undef EXIT_FUNCTION

/**************************************************************/
/*                                                            */
/* Routine  : Execute                                         */
/*                                                            */
/* Description  : Execution routine for a sample exit. This   */
/*                is called ....                              */
/*                From the inbound server immediately after   */
/*                the MQGET from the input queue              */
/*                From the outbound server immediately before */
/*                the MQPUT to the destination queue.         */
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure.                                 */
/*               pExitInput - pointer to an SMQEXIT_INPUT     */
/*                 structure. The structure contains details  */
/*                 about the data to be processed and ptrs    */
/*                 to the data. This structure is input only  */
/*                 and should not be changed by this routine. */
/*               pExitOutput - pointer to an SMQEXIT_OUTPUT   */
/*                 structure. This is used to contain info.   */
/*                 about the converted data.                  */
/*                                                            */
/**************************************************************/
#define EXIT_FUNCTION "Execute     "
void MQENTRY Execute(PSMQEXIT_SETUP pExitSetup,
                     PSMQEXIT_INPUT pExitInput,
                     PSMQEXIT_OUTPUT pExitOutput)
{
  pTransLog pLogData ;
  MQLONG    CompCode ;
  MQLONG    ReasonCode ;
  pLogBuffer pTBuffer ;

  /************************************************************/
  /* Establish addressability to our private data.            */
  /************************************************************/
  pLogData = (pTransLog)(pExitSetup -> ExitHandle) ;
  pTBuffer = pLogData -> pBuffer ;

  /************************************************************/
  /* Check the direction indicator which we saved in our      */
  /* own control block and call the appropiate subroutine.    */
  /************************************************************/
  if (pLogData -> Direction == SMQEXIT_INBOUND)
  {
    transInbound(pExitSetup,
                 pExitInput,
                 pExitOutput) ;
  }
  else
  {
    transOutbound(pExitSetup,
                 pExitInput,
                 pExitOutput) ;
  }

  /************************************************************/
  /* Write a message to our log queue.                        */
  /************************************************************/

  memcpy(pTBuffer -> Phase,
         EXIT_FUNCTION,
             12) ;
  pTBuffer -> Result = pExitOutput -> Result ;
  MQPUT(pLogData -> QMgrHandle,
        pLogData -> LogQHandle,
        pLogData -> pLogMQMD,
            pLogData -> pLogMQPMO,
                sizeof(LogBuffer),
                pLogData -> pBuffer,
            &CompCode,
                &ReasonCode) ;

  return ;
}
#undef EXIT_FUNCTION


/**************************************************************
 *
 *                                              transInbound
 *
 **************************************************************/
void MQENTRY transInbound(PSMQEXIT_SETUP pExitSetup,
                          PSMQEXIT_INPUT pExitInput,
                          PSMQEXIT_OUTPUT pExitOutput)
{
  pTransLog pLogData ;
  PMQSAPH pHeader ;
  PMQVOID pTemp ;

  /* ---------------------------------- *\
   | Declare the variables that we will |
   | be using to call the RunMap func.  |
  \* ---------------------------------- */
  EXITPARAM ep;
  DWORD nCmdLineLen = 0;       // command line length

  /* ---------------------------------------------- *\
   | Establish addressability to our private data.  |
  \* ---------------------------------------------- */
  pLogData = (pTransLog)(pExitSetup -> ExitHandle) ;


  /* ---------------------------------------------- *\
   | Construct the comand line that we will be      |
   | passing to Mercator. Note: Better change the   |
   | map name below for your system!                |
  \* ---------------------------------------------- */
  ep.lpDataToApp = malloc( pExitInput->MessageDataLen + 300 );
  if( NULL == ep.lpDataToApp )
          return;              // Failure

  sprintf( ep.lpDataToApp, "App2IDoc.MMC -ie1s%d ",
               pExitInput->MessageDataLen);
  nCmdLineLen = strlen ( ep.lpDataToApp );

  memcpy( (void *)(ep.lpDataToApp + nCmdLineLen),
                  pExitInput->MessageData,
                  pExitInput->MessageDataLen );
  nCmdLineLen = nCmdLineLen + pExitInput->MessageDataLen;

  {
          char * lpszCmdEnd = " -oe1\0";
          memcpy( (void *)(ep.lpDataToApp + nCmdLineLen),
                          lpszCmdEnd,
                          strlen( lpszCmdEnd ) + 1 );
  }

  /* ---------------------------------------------- *\
   | Now that we finally got the command line       |
   | together just fire off the map.                |
   |                                                |
   | Upon return of the mapping, your code should   |
   | probably check the nReturn field. If if fails  |
   | you may want to return immediately or even go  |
   | through retries. Just something to think about.|
  \* ---------------------------------------------- */
  RunMap( &ep );


  /* ---------------------------------------------- *\
   | Now set up a 'default' MQSAPH at the start of  |
   | the data. This also means that we have to set  |
   | up the Header pointer appropriately. It is     |
   | noted in the sample user exit (by IBM) that the|
   | default MQSAPH might have "nulls" for some of  |
   | the character fields so we have to set some of |
   | the fields to blanks in order to pick up the   |
   | 'default' values for some of the parameters    |
   | from the ini file                              |
  \* ---------------------------------------------- */
  pHeader = (PMQSAPH)malloc( sizeof(MQSAPH) + ep.dwFromLen );
  if ( NULL == pHeader )
  {
    pExitOutput -> Result = SMQRC_CONVERT_FAIL ;
    return ;
  }
  else
  {
  /* ---------------------------------------------- *\
   | Okay, we got the memory we need in order to    |
   | store the mapped data and the Link Header. Now |
   | we have to put the default Link header in front|
   | of the mapped data and pass it all back to     |
   | MQLink.                                        |
  \* ---------------------------------------------- */
        memcpy(pHeader,
               &DefHeader,
               sizeof(MQSAPH)) ;
        pExitOutput -> Header = pHeader ;
        pTemp = &(pHeader -> Client) ;
        memset(pTemp,
                   ' ',
                   sizeof(MQSAPH) - 32) ;

  /* ---------------------------------------------- *\
   | Now that we have constructed a Link header we  |
   | are ready to copy the mapped data into the     |
   | 'data' section of the message data, i.e. The   |
   | space after the Link header.                   |
  \* ---------------------------------------------- */
        pExitOutput->MessageData = (char *)pHeader + sizeof(MQSAPH);
        memcpy ( pExitOutput->MessageData,
                     ep.lpDataFromApp,
                         ep.dwFromLen);
        pExitOutput->MessageDataLen = ep.dwFromLen;
        pExitOutput->MessageDesc        = pExitInput->MessageDesc;

  /* ---------------------------------------------- *\
   | Free the LPEXITPARAM structure before we go.   |
   | Note: It would be wise to free the lpDataToApp |
   | as soon as we get back from the RunMap call    |
   | it is no longer useful and the amount of mem.  |
   | tied up is great. Also you would want to do    |
   | this cleaning before any of the return's.....  |
   | but you already knew that.                     |
  \* ---------------------------------------------- */
        GlobalFree( ep.lpDataFromApp );
        free( ep.lpDataToApp );

    /* ------------------------------------------------ *\
     | This was part of the original sample source....  |
     |                                                  |
     | Remind ourselves that we must free the storage   |
     | acquired above in the return routine.            |
    \* ------------------------------------------------ */
    pLogData->Flags             = TRANS_FREE_HEADER ;
        pExitOutput->Result = SMQRC_CONVERT_OK ;
  }

  return ;
}

void MQENTRY transOutbound(PSMQEXIT_SETUP pExitSetup,
                           PSMQEXIT_INPUT pExitInput,
                           PSMQEXIT_OUTPUT pExitOutput)
{
  pTransLog             pLogData ;
//  PMQVOID             pOutputData ;
  PMQMD                 ptMD ;

  /* ---------------------------------- *\
   | Declare the variables that we will |
   | be using to call the RunMap func.  |
  \* ---------------------------------- */
  EXITPARAM ep;
  char szCommandBegin[100];
  char szCommandEnd[] = "-oe1\0";
  DWORD nCommandBeginLen = 0;
  DWORD nCommandEndLen   = 0;

  /* ---------------------------------------------- *\
   | Establish addressability to our private data.  |
  \* ---------------------------------------------- */
  pLogData = (pTransLog)(pExitSetup -> ExitHandle) ;

  /* ---------------------------------------------- *\
   | For this particular outbound transformation we |
   | will be assuming that the information will     |
   | come in as IDOC. As for the output we will not |
   | be creating a Link header because it is going  |
   | to be assumed that the destination is not      |
   | another R/3 system or any system that would    |
   | need the Link header.                          |
   |                                                |
   | It is reasonable to check that the IDoc given  |
   | is one that we will map but for this sample we |
   | aren't checking.                               |
  \* ---------------------------------------------- */

  /* ---------------------------------------------- *\
   | Construct the comand line that we will be      |
   | passing to Mercator. Note: Better change the   |
   | map name below for your system!                |
  \* ---------------------------------------------- */
  sprintf (szCommandBegin, "\\App2IDoc.MMC -ie1s%d ",
               pExitInput->MessageDataLen);
  nCommandBeginLen = strlen (szCommandBegin);
  nCommandEndLen   = strlen (szCommandEnd);
  ep.dwToLen = nCommandBeginLen +
                                  nCommandEndLen   +
                                  pExitInput->MessageDataLen;
  ep.lpDataToApp = malloc (ep.dwToLen + 1);

  /* ---------------------------------------------- *\
   | Now that we have all the command line pieces   |
   | we have to put them all in the proper field... |
  \* ---------------------------------------------- */
  memcpy (  (void * )ep.dwToLen,
                        szCommandBegin,
                        nCommandBeginLen);
  memcpy ( (void * )(ep.dwToLen+nCommandBeginLen),
                        pExitInput->MessageData,
                        pExitInput->MessageDataLen);
  memcpy ( (void * )(ep.dwToLen+nCommandBeginLen+pExitInput->MessageDataLen),
                        szCommandEnd,
                        nCommandEndLen + 1);

  /* ---------------------------------------------- *\
   | Now that we finally got the command line       |
   | together just fire off the map.                |
   |                                                |
   | Upon return of the mapping, your code should   |
   | probably check the nReturn field. If if fails  |
   | you may want to return immediately or even go  |
   | through retries. Just something to think about.|
  \* ---------------------------------------------- */
  RunMap( &ep );

  /* ---------------------------------------------- *\
   | Now we need to copy the mapped data into the   |
   | proper parameter for MQLink to deal with. We   |
   | could just simply make this assignment:        |
   |    pExitOutput->Message = ep.lpDataFromApp;    |
   | but the memory should be freed with a call to  |
   | GlobalFree() and we don't know if that is how  |
   | MQLink would free the memory. Therefore we are |
   | going to do it ourselves.                      |
  \* ---------------------------------------------- */
  pExitOutput->MessageData    = (char *)malloc (ep.dwFromLen);
  memcpy ( pExitOutput->MessageData,
                   ep.lpDataFromApp,
                   ep.dwFromLen);
  pExitOutput->MessageDataLen = ep.dwFromLen;

  ptMD = (void *)malloc(sizeof(MQMD));

  /* -------------------------------------- *\
   | Remind ourselves to free the ptMD var. |
  \* -------------------------------------- */
   pLogData -> Flags = TRANS_FREE_MQMD ;

  /* -------------------------------------- *\
   | Now since we are creating a new format |
   | we have to specify that format in the  |
   | Message Descriptor. i.e. the output is |
   | not in IDoc format, it's now something |
   | else...FMT_APPLICATION.                |
   | You should change the value according  |
   | the type of Format that your exit will |
   | be creating                            |
  \* -------------------------------------- */
  memcpy(ptMD,
             pExitInput -> MessageDesc,
                 sizeof(MQMD)) ;
  memcpy(ptMD -> Format,
             FMT_APPLICATION,
                 8) ;
  /* ---------------------------------------------- *\
   | Free the LPEXITPARAM structure before we go.   |
   | Note: It would be wise to free the lpDataToApp |
   | as soon as we get back from the RunMap call    |
   | it is no longer useful and the amount of mem.  |
   | tied up is great. Also you would want to do    |
   | this cleaning before any of the return's.....  |
   | but you already knew that.                     |
  \* ---------------------------------------------- */
  GlobalFree (ep.lpDataFromApp);
  free (ep.lpDataToApp);


  pExitOutput->MessageDesc = ptMD ;
  pExitOutput -> Result = SMQRC_CONVERT_OK ;

  return ;
}


/**************************************************************/
/*                                                            */
/* Routine  : Return                                          */
/*                                                            */
/* Description  : Return routine for sample exit. This        */
/*                routine is called when the server has       */
/*                completed processing the data returned by   */
/*                the execute function. Its function is to    */
/*                cleanup, if necessary, any resources        */
/*                acquired by the execute function. For our   */
/*                purposes here we check the flags in our     */
/*                control structure to determine if execute   */
/*                acquired any buffers. If so we free the     */
/*                storage.                                    */
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure. This is the structure as        */
/*                 returned by the previous call to an exit   */
/*                 routine.                                   */
/*               pExitOutput - pointer to an SMQEXIT_OUTPUT   */
/*                 structure. This is the structure as        */
/*                 returned by the previous call to the       */
/*                 execute function.                          */
/*                                                            */
/* Design  : TBC                                              */
/*                                                            */
/**************************************************************/
#define EXIT_FUNCTION "Return      "
void MQENTRY Return(PSMQEXIT_SETUP pExitSetup,
                    PSMQEXIT_OUTPUT pExitOutput)
{
  pTransLog pLogData ;
  MQLONG    CompCode ;
  MQLONG    ReasonCode ;
  pLogBuffer pTBuffer ;

  /************************************************************/
  /* Set OK as the default result.                            */
  /************************************************************/
  pExitOutput -> Result = SMQRC_OK ;

  /************************************************************/
  /* In this sample we must free any buffers acquired in the  */
  /* execute routine. First we get addressabilty to our       */
  /* control structure. Then we test the Flags field in the   */
  /* control structure to check which buffers we must free.   */
  /* After freeing the necessary storage we turn off all the  */
  /* storage-related bits.                                    */
  /************************************************************/
  pLogData = (pTransLog)(pExitSetup -> ExitHandle) ;
  switch (pLogData -> Flags)
  {
    case TRANS_FREE_HEADER:
          free(pExitOutput -> Header) ;
          break ;
    case TRANS_FREE_MQMD:
          free(pExitOutput -> MessageDesc) ;
      break ;
    case TRANS_FREE_MESSAGE:
          free(pExitOutput -> MessageData) ;
      break ;
    default:
          break ;
  }
  pLogData -> Flags = 0 ;

  /************************************************************/
  /* Write a message to our log queue.                        */
  /************************************************************/
  pTBuffer = pLogData -> pBuffer ;
  memcpy(pTBuffer -> Phase,
         EXIT_FUNCTION,
             12) ;
  pTBuffer -> Result = pExitOutput -> Result ;
  MQPUT(pLogData -> QMgrHandle,
        pLogData -> LogQHandle,
        pLogData -> pLogMQMD,
            pLogData -> pLogMQPMO,
                sizeof(LogBuffer),
                pLogData -> pBuffer,
        &CompCode,
            &ReasonCode) ;

  /************************************************************/
  /* Now return to the exit handler.                          */
  /************************************************************/
  return ;
}
#undef EXIT_FUNCTION

